# -*- coding: utf-8 -*-

# Copyright (C) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

import sys
import xml.etree.ElementTree as ET
from xml.dom import minidom

def prettify(elem):
    #Return a pretty-printed XML string for the Element.
    rough_string = ET.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="    ")

argv = sys.argv
if len(argv) < 2:
    print("Usage: " + argv[0] + " input.xml [output.xml]")
    sys.exit()
inputxml = argv[1]
outputxml = argv[2] if len(argv) >= 3 else "output.xml"

# Load xml and validation
orgTree = ET.parse(inputxml)
orgRoot = orgTree.getroot()
if orgRoot.tag != "hark_xml":
    print("Input file is not HARK xml!")
    sys.exit()
orgConfig = orgRoot.find("./hark_config")
if orgConfig is None or orgConfig.get("major") != "2" or orgConfig.get("minor") != "0":
    print("Unspported HARK xml format!")
    sys.exit()
srcType = orgConfig.get("id")
if srcType is None or srcType != "MicArrayLocation":
    print("Unspported source id!")
    sys.exit()

""" Process conversion
"""
newRoot = ET.Element("hark_xml", {'version': '1.3'})

# position
positions = ET.SubElement(newRoot, "positions", {'type': 'microphone', 'coordinate': 'cartesian'})

def asc_cmp(x, y):
    return int(x.findtext("./MicIndex", "0")) - int(y.findtext("./MicIndex", "0"))
micList = orgConfig.findall("./Mics/Mic")
micList.sort(asc_cmp)
id = 0
for e in micList:
    while id < int(e.findtext("./MicIndex", "")):
        # insert dummy microphone channel
        ET.SubElement(positions, "position", {'x': "0.0000", 'y': "0.0000", 'z': "0.0000", 'id': str(id)})
        id += 1
        if id >= 64:
            print("Microphone channel too large!")
            sys.exit()
    pos = e.find("Position")
    if pos is not None:
        if pos.get("type") != "cartesian":
            print("Unsupported position coordinate!")
            sys.exit()
        pos_x = (pos.findtext("X", ""))
        pos_y = (pos.findtext("Y", ""))
        pos_z = (pos.findtext("Z", ""))

        ET.SubElement(positions, "position", {'x': pos_x, 'y': pos_y, 'z': pos_z, 'id': str(id)})
    id += 1

# Write
with open(outputxml, "wb") as f:
    f.write(prettify(newRoot))
