/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef WIN32
#define _GNU_SOURCE
#endif
//#include <iostream>
//#include <complex>
#include <complex.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdbool.h>
#include <math.h>

#include "harktool.h"

/*-----------------------------------------------------------------------------*
 //
 *-----------------------------------------------------------------------------*/
/*
 extern long get_flt_size(const char* fname);
 extern int read_flt(const char* fname, double* dBuf, size_t file_size, double* signal_max);
 extern int write_flt(const char* fname, const double* dBuf, size_t count);
 extern int readwavfile(char *fname, double ***wave, int *length, int *ch, int *samplerate);
 extern int writewavfile(char *fname, double **wave, int length, int ch, int samplerate);
 extern void harktool_calcDoubleFFT( COMPLEX_DOUBLE  *a , COMPLEX_DOUBLE  *A , int data_size );
 */
void ifft(double *in_buf_re, double *in_buf_im, double *out_buf_re, double *out_buf_im, int data_size);
void fft(double *in_buf, double *out_buf_re, double *out_buf_im, int data_size);

//#define DEBUG 0

//#define IFDEBUG(statement) statement

#ifdef GPL
#include <fftw.h>
#endif

#ifdef HARKLICENSE2
#include "fft.hpp"
#endif

#ifdef DEBUG
#define IFDEBUG(statement) statement
#else
#define IFDEBUG(statement)
#endif

#ifdef cplusplus
extern "C" {
#endif
void fft(double *in_buf, double *out_buf_re, double *out_buf_im, int data_size);
void ifft(double *in_buf_re, double *in_buf_im, double *out_buf_re, double *out_buf_im, int data_size);
int corr_on_fft(double *x, int x_size, double *y, int y_size, int start_pos, int end_pos);

#ifdef cplusplus
}
#endif

#define DISTANCE_OFFSET (-50) // margin for speaker-mic distance.

//void usage()
//{
//    printf("./calcSysImp [option] inputfile outimpfilehead \n");
//    printf("calculate impluse from TSP wave\n");
//    printf("(fix time: 2013/10/30, compile time: %s %s)\n",__DATE__ ,__TIME__);
//    printf("\n");
//    printf("DESCRIPTION\n");
//    printf("\tinputfile   : 16bit or 24bit , 8ch or 10ch or 16ch , 16kHz wave format\n");
//    printf("\t              48KHz is not available yet.\n");
//    printf("\timpfilehead : its extention is k.flt where k is ch number.\n");
//    printf("\t              each imp file is 32bit float mono 1-ch raw data\n");
//    printf("\t  eg.) when outimpfilehead = 'a', file names are 'a_ch1.flt', 'a_ch2.flt', ...\n");
//    printf("\n");
//    printf("[option]\n");
//    printf("\t--tsporigflt file_name   : playing tsp file(float-data) when recording tsp wave.\n");
//    printf("\t--tsporigwav file_name   : playing tsp file(PCM-wave) when recording tsp wave.\n");
//    printf("\t--tsporigchannel N       : select use cahnnel of tsp origin file(PCM-wave).\n");
//    printf("\t--sigmax N               : set optional signal max for tsp original file\n");
//    printf("\t                         (not working on using --tsporigwav.)\n");
//    printf("\t--samplebyframe N        : sample by one frame(default = 16384) \n");
//    printf("\t--syncaddfreq N          : frequency of synchronous addition(default = 16) \n");
//    printf("\t--syncaddoffset N        : start position of synchronous addition(default = 16384) \n");
//    printf("\t--searchoffset N,N       : using searched start position.(search start pos, search end pos is required.) \n");
//    printf("\t--outchannellist N,N,... : select calulating channel.(first channel is ch0.)\n");
//    printf("\t--enableHPF Lfreq,Hfreq  : enable HighPassFilter.\n");
//
//    exit(-1);
//    /*./calcSysImp ../TSP000.wav ../impulse/D000 */
//
//}

/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
int harktool_calcSysImp(enum SoundType tspOrgType, const char* tspOrgFile, int tspOrgChannel, int signalMax,
        int cfgSamplingRate, int samplesByFrame, int syncAddFreq, int syncAddOffset, int searchOffsetFrom/* =-1 */,
        int searchOffsetTo/* =-1 */, const bool* outChannels, int chOrigin, const char* inputWaveFile,
        const char* outputFltFile) {
    // print arguments
    char* msg = NULL;
    cat_asprintf(&msg, "\nharktool_calcSysImp   tspOrgType       [%s]\n", (tspOrgType == RAW) ? "RAW" : "WAV");
    cat_asprintf(&msg, "                    , tspOrigWav       [%s]\n", tspOrgFile); // "/path/to/16384.little_endian.wav"
    cat_asprintf(&msg, "                    , tspOrigChannel   [%d]\n", tspOrgChannel); // ch0 only
    cat_asprintf(&msg, "                    , signalMax        [%d]\n", signalMax); // 32768
    cat_asprintf(&msg, "                    , samplingRate     [%d]\n", cfgSamplingRate); // 16000
    cat_asprintf(&msg, "                    , tspLength        [%d]\n", samplesByFrame); // 16384
    cat_asprintf(&msg, "                    , syncAddFreq      [%d]\n", syncAddFreq); // 16
    cat_asprintf(&msg, "                    , syncAddOffset    [%d]\n", syncAddOffset); // 16384
    cat_asprintf(&msg, "                    , searchOffsetFrom [%d]\n", searchOffsetFrom); // 0
    cat_asprintf(&msg, "                    , searchOffsetTo   [%d]\n", searchOffsetTo); // 32
    cat_asprintf(&msg, "                    , outChannelList   [");
    bool printedCh = false;
    for (int i = 0; i < MAX_CH; i++) {
        if (outChannels[i]) {
            if (printedCh)
                cat_asprintf(&msg, ",");
            printedCh = true;
            cat_asprintf(&msg, "%d", i);
        }
    }
    cat_asprintf(&msg, "]\n");
    cat_asprintf(&msg, "                    , channelOrigin    [%d]\n", chOrigin);
    cat_asprintf(&msg, "                    , inputWav         [%s]\n", inputWaveFile); // "/path/to/src.wav"
    cat_asprintf(&msg, "                    , outputPath       [%s]\n", outputFltFile); // "/path/to/src.wav/float"
    harkio_Log_printf(harkio_Log_Info, "%s", msg);
    free(msg);

    // assert(searchOffsetFrom >= 0)
    // assert(searchOffsetFrom <= searchOffsetTo)
    double tsp_adjust_ratio = signalMax;

    bool out_channel_set_flag = false;
    for (int i = 0; i < MAX_CH; i++) {
        if (!outChannels[i]) {
            out_channel_set_flag = true;
            break;
        }
    }

    bool search_offset_flag = false;
    if (searchOffsetFrom >= 0)
        search_offset_flag = true;

    bool enable_HPF_flg = 0;
    int enable_HPF_cutLow_freq = 0;
    int enable_HPF_cutHigh_freq = 0;

    double *tsp_wave_buf = NULL;
    double **tsp_wave;
    int tsp_length;

    int h_size = 4096;
    int ch_idx = 0;
    int out_ch_idx = 0;

    if (search_offset_flag && !out_channel_set_flag) { //search_offset needs out_channel_list
        harkio_Log_printf(harkio_Log_Error, "searchOffset needs outChannels for calculation.");
    }

    // TSP波形読み込み
//    double tsp_fft_re[samplesByFrame];
//    double tsp_fft_im[samplesByFrame];
    COMPLEX_DOUBLE tsp_fft[samplesByFrame];
    int frame_cnt;
    double dummy;

    if (tspOrgType == RAW) {
        if (cfgSamplingRate != 16000) {
            harkio_Log_printf(harkio_Log_Error, "Config SamplingRate shuold be 16000Hz. (%d)", cfgSamplingRate);
            return EXIT_FAILURE;
        }
        int tsp_file_size = get_flt_size(tspOrgFile);
        tsp_length = tsp_file_size / sizeof(float);

        tsp_wave_buf = (double *) malloc(tsp_file_size / sizeof(float) * sizeof(double));

        if (signalMax > 0) {
            if (read_flt(tspOrgFile, tsp_wave_buf, tsp_file_size, &dummy) == -1) {
                harkio_Log_printf(harkio_Log_Error, "tsp origin file is not read.");
                return EXIT_FAILURE;
            }
        } else {
            if (read_flt(tspOrgFile, tsp_wave_buf, tsp_file_size, &tsp_adjust_ratio) == -1) {
                harkio_Log_printf(harkio_Log_Error, "tsp origin file is not read.");
                return EXIT_FAILURE;
            }
        }

        for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
            tsp_wave_buf[frame_cnt] /= tsp_adjust_ratio;
        }

//        fft(tsp_wave_buf, tsp_fft_re, tsp_fft_im, samplesByFrame);
        harktool_FFTImpl_dRealFFT(tsp_wave_buf, tsp_fft, samplesByFrame);
    } else if (tspOrgType == WAV) {
        int tsp_ch;
        int tsp_samplerate;

        if (readwavfile(tspOrgFile, &tsp_wave, &tsp_length, &tsp_ch, &tsp_samplerate) == -1) {
            harkio_Log_printf(harkio_Log_Error, "wavefile(%s) is not found.", tspOrgFile);
            return EXIT_FAILURE;
        }

        if (cfgSamplingRate != tsp_samplerate) {
            harkio_Log_printf(harkio_Log_Error, "Config SamplingRate (%d) differ from Wav(%s) SamplingRate (%d)",
                    cfgSamplingRate, tspOrgFile, tsp_samplerate);
            return EXIT_FAILURE;
        }

//        fft(tsp_wave[tspOrgChannel], tsp_fft_re, tsp_fft_im, samplesByFrame);
        harktool_FFTImpl_dRealFFT(tsp_wave[tspOrgChannel], tsp_fft, samplesByFrame);

    } else {
        harkio_Log_printf(harkio_Log_Error, "Error: incorrect tsp_org_type.");
        return EXIT_FAILURE;  // exit in funcition.
    }

    IFDEBUG(fprintf(debug_fd, "/////TSP FFT/////\n");
            for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
                fprintf(debug_fd,"%f+%fi\n", creal(tsp_fft[frame_cnt]), cimag(tsp_fft[frame_cnt]));
            }
            fprintf(debug_fd, "\n\n\n\n");
    )

    IFDEBUG(fflush(debug_fd));

    // マイク録音読み込み
    double **iwave;
    int ilength;
    int input_ch;
    int isamplerate;

    if (readwavfile(inputWaveFile, &iwave, &ilength, &input_ch, &isamplerate) == -1) {
        harkio_Log_printf(harkio_Log_Error, "wavefile(%s) is not found.", inputWaveFile);
        return EXIT_FAILURE;
    }

    if (((syncAddFreq * samplesByFrame) + syncAddOffset) > ilength) {
        harkio_Log_printf(harkio_Log_Error, "Wave length(%d) is short when syncAddFreq = %d.", ilength, syncAddFreq);
        return EXIT_FAILURE;
    }

    IFDEBUG(
            fprintf(debug_fd, "/////INPUT WAVE/////\n");
            for (frame_cnt = 0; frame_cnt < /*samplesByFrame*/1024; frame_cnt++) {
                fprintf(debug_fd, "%f\n", iwave[0][frame_cnt]);
            }
            fprintf(debug_fd, "\n\n\n\n");
    )

    IFDEBUG(fclose(debug_fd)); IFDEBUG(debug_fd = fopen("debug.txt", "a+t"));

    // インパルス応答計算
    harkio_Log_printf(harkio_Log_Info, "calc_impulse...");
    double xsum[samplesByFrame];
    double xmean[samplesByFrame];
//    double xfft_re[samplesByFrame];
//    double xfft_im[samplesByFrame];
//    double h_wave_buf_re[samplesByFrame];
//    double h_wave_buf_im[samplesByFrame];
//    double hfft_re[samplesByFrame];
//    double hfft_im[samplesByFrame];
//    double h_wave_short_buf_re[h_size];
//    double h_wave_short_buf_im[h_size];
//    double hhpf_re[h_size];
//    double hhpf_im[h_size];
    COMPLEX_DOUBLE xfft[samplesByFrame];
    COMPLEX_DOUBLE h_wave_buf[samplesByFrame];
    COMPLEX_DOUBLE hfft[samplesByFrame];
    COMPLEX_DOUBLE h_wave_short_buf[h_size];
    COMPLEX_DOUBLE hhpf[h_size];
    double hpf_coef[h_size];

    if (search_offset_flag) {
        harkio_Log_printf(harkio_Log_Info, "search_offset...");
        int calculated_offset, calculated_first_offset = searchOffsetTo;
        //オフセット推定処理
        for (ch_idx = 0; ch_idx < input_ch; ch_idx++) {
            if (!outChannels[ch_idx]) {
                continue;
            }

            if (searchOffsetTo > ilength) {
                searchOffsetTo = ilength;
            }

            // 探索範囲がTSPのサイズより狭かった場合、TSPのサイズまで広げる
            if ((searchOffsetTo - searchOffsetFrom) < tsp_length) {
                if ((searchOffsetFrom - tsp_length) >= 0) {
                    searchOffsetFrom -= tsp_length;
                    harkio_Log_printf(harkio_Log_Info,
                            "Search length is under tsp size. So changing search start pos = %d.", searchOffsetFrom);
                } else if ((searchOffsetTo + tsp_length) <= ilength) {
                    searchOffsetTo += tsp_length;
                    harkio_Log_printf(harkio_Log_Info,
                            "Search length is under tsp size. So changing search end pos = %d.", searchOffsetTo);
                } else {
                    harkio_Log_printf(harkio_Log_Error,
                            "Search length is under tsp size. But there is not enough area on impulse.");
                    return EXIT_FAILURE;
                }
            }

            if (tspOrgType == RAW) {
                //syncAddOffset = corr(iwave[0], ilength, tsp_wave_buf, tsp_file_size, searchOffsetFrom, searchOffsetTo);
                calculated_offset = corr_on_fft(iwave[ch_idx], ilength, tsp_wave_buf, tsp_length, searchOffsetFrom,
                        searchOffsetTo);
            } else {
                //syncAddOffset = corr(iwave[0], ilength, tsp_wave[tsp_org_channel], tsp_length, searchOffsetFrom, searchOffsetTo);
                calculated_offset = corr_on_fft(iwave[ch_idx], ilength, tsp_wave[tspOrgChannel], tsp_length,
                        searchOffsetFrom, searchOffsetTo);
            }
            if (calculated_offset < calculated_first_offset) {
                calculated_first_offset = calculated_offset;
            }
        }
        harkio_Log_printf(harkio_Log_Info, "calculated_first_offset=%d.", calculated_first_offset);
        syncAddOffset = calculated_first_offset;
    }

    for (ch_idx = 0, out_ch_idx = 0; ch_idx < input_ch; ch_idx++) {
        int frame_idx;

        if (outChannels[ch_idx] == 0) {
            continue;
        }

        for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
            xsum[frame_cnt] = 0;

            for (frame_idx = 0; frame_idx < syncAddFreq; frame_idx++) {
                xsum[frame_cnt] += iwave[ch_idx][frame_idx * samplesByFrame + frame_cnt + syncAddOffset];

            }
        }

        for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
            xmean[frame_cnt] = xsum[frame_cnt] / syncAddFreq;
        }

//        fft(xmean, xfft_re, xfft_im, samplesByFrame);
        harktool_FFTImpl_dRealFFT(xmean, xfft, samplesByFrame);

        IFDEBUG(
                fprintf(debug_fd, "/////X FFT/////\n");
                for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
                    fprintf(debug_fd, "%f+%fi\n", creal(xfft[frame_cnt]), cimag(xfft[frame_cnt]));
                }
                fprintf(debug_fd, "\n\n\n\n");
                fprintf(debug_fd, "fft end...\n");
        )

        for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
//            double denomi = tsp_fft_re[frame_cnt] * tsp_fft_re[frame_cnt] + tsp_fft_im[frame_cnt] * tsp_fft_im[frame_cnt];
            double denomi = tsp_fft[frame_cnt] * conj(tsp_fft[frame_cnt]);
//            hfft_re[frame_cnt] = (xfft_re[frame_cnt] * tsp_fft_re[frame_cnt] + xfft_im[frame_cnt] * tsp_fft_im[frame_cnt])
//                                 / denomi;
//            hfft_im[frame_cnt] = (tsp_fft_re[frame_cnt] * xfft_im[frame_cnt] - xfft_re[frame_cnt] * tsp_fft_im[frame_cnt])
//                                 / denomi;
            hfft[frame_cnt] = xfft[frame_cnt] * conj(tsp_fft[frame_cnt]) / denomi;
        }

        IFDEBUG(
                fprintf(debug_fd, "/////H FFT/////\n");

                for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
                    fprintf(debug_fd, "%f+%fi\n", creal(hfft[frame_cnt]), cimag(hfft[frame_cnt]));
                }
                fprintf(debug_fd, "\n\n\n\n");
        )

//        ifft(hfft_re, hfft_im, h_wave_buf_re, h_wave_buf_im, samplesByFrame);
        harktool_FFTImpl_dFFTInv(hfft, h_wave_buf, samplesByFrame);

        for (frame_cnt = 0; frame_cnt < samplesByFrame; frame_cnt++) {
//            h_wave_buf_re[frame_cnt] /= samplesByFrame;
//            h_wave_buf_im[frame_cnt] /= samplesByFrame;
            h_wave_buf[frame_cnt] /= samplesByFrame;
        }

        // 初期フレームだけ切り出し
        for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
//            h_wave_short_buf_re[frame_cnt] = h_wave_buf_re[frame_cnt];
//            h_wave_short_buf_im[frame_cnt] = h_wave_buf_im[frame_cnt];
            h_wave_short_buf[frame_cnt] = h_wave_buf[frame_cnt];
        }

        IFDEBUG(
                fprintf(debug_fd, "/////h short wave/////\n");
                for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                    fprintf(debug_fd, "%f+%fi\n", creal(h_wave_short_buf[frame_cnt]), cimag(h_wave_short_buf[frame_cnt]));
                }
                fprintf(debug_fd, "\n\n\n\n");
        )

        if (enable_HPF_flg) {

//            fft(h_wave_short_buf_re, hhpf_re, hhpf_im, h_size);
            double h_wave_short_buf_re[h_size];
            for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                h_wave_short_buf_re[frame_cnt] = creal(h_wave_short_buf[frame_cnt]);
            }
            harktool_FFTImpl_dRealFFT(h_wave_short_buf_re, hhpf, h_size);

            IFDEBUG(
                    fprintf(debug_fd, "/////before HPF/////\n");
                    for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                        fprintf(debug_fd, "%f+%fi\n", creal(hhpf[frame_cnt]), cimag(hhpf[frame_cnt]));
                    }
                    fprintf(debug_fd, "\n\n\n\n");
            )

            // 低周波カット
            double cut_f1 = enable_HPF_cutLow_freq;
            double cut_f2 = enable_HPF_cutHigh_freq;
            double fs = 16000;
            int cut1 = round(cut_f1 * h_size / fs);
            int cut2 = round(cut_f2 * h_size / fs);
            double cut_d = cut2 - cut1;

            for (frame_cnt = 0; frame_cnt < h_size / 2 + 1; frame_cnt++) {
                if (frame_cnt < cut1) {
                    hpf_coef[frame_cnt] = 0;
                    hpf_coef[h_size - frame_cnt - 1] = 0;
                } else if (frame_cnt < cut2) {
                    hpf_coef[frame_cnt] = (frame_cnt + 1 - cut1) / cut_d;
                    hpf_coef[h_size - frame_cnt] = (frame_cnt + 1 - cut1) / cut_d;
                } else {
                    hpf_coef[frame_cnt] = 1;
                    hpf_coef[h_size - frame_cnt] = 1;
                }
            }

            IFDEBUG(
                    fprintf(debug_fd, "/////HPF COEF/////\n");
                    for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                        fprintf(debug_fd, "%f\n", hpf_coef[frame_cnt]);
                    }
                    fprintf(debug_fd, "cutHz=%f:%f", cut_f1, cut_f2);
                    fprintf(debug_fd, "\n\n\n\n");
            )

            for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
//                hhpf_re[frame_cnt] = hhpf_re[frame_cnt] * hpf_coef[frame_cnt];
//                hhpf_im[frame_cnt] = hhpf_im[frame_cnt] * hpf_coef[frame_cnt];
                hhpf[frame_cnt] *= hpf_coef[frame_cnt];
            }

            IFDEBUG(
                    fprintf(debug_fd, "/////after HPF/////\n");
                    for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                        fprintf(debug_fd, "%f+%fi\n", creal(hhpf[frame_cnt]), cimag(hhpf[frame_cnt]));
                    }
                    fprintf(debug_fd, "\n\n\n\n");
            )

//            ifft(hhpf_re, hhpf_im, h_wave_short_buf_re, h_wave_short_buf_im, h_size);
            harktool_FFTImpl_dFFTInv(hhpf, h_wave_short_buf, h_size);

            for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
//                h_wave_short_buf_re[frame_cnt] /= h_size;
//                h_wave_short_buf_im[frame_cnt] /= h_size;
                h_wave_short_buf[frame_cnt] /= h_size;
            }

        IFDEBUG(
                fprintf(debug_fd, "/////short wave after HPF/////\n");
                for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
                    fprintf(debug_fd, "%f+%fi\n", creal(h_wave_short_buf[frame_cnt]), cimag(h_wave_short_buf[frame_cnt]));
                }
                fprintf(debug_fd, "\n\n\n\n");
        )

    }

    char* o_wave_name = NULL;
    asprintf(&o_wave_name, "%s_ch%d.flt", outputFltFile, out_ch_idx + chOrigin);

    //int write_flt(const char* fname, const double* dBuf, size_t count)
    double h_wave_short_buf_re[h_size];
    for (frame_cnt = 0; frame_cnt < h_size; frame_cnt++) {
        h_wave_short_buf_re[frame_cnt] = creal(h_wave_short_buf[frame_cnt]);
    }
    if (write_flt(o_wave_name, h_wave_short_buf_re, h_size) == -1) {
        harkio_Log_printf(harkio_Log_Error, "File open failed: %s", o_wave_name);
        if (o_wave_name)
            free(o_wave_name);
        return EXIT_FAILURE;
    }
    free(o_wave_name);

    out_ch_idx++;
}

harkio_Log_printf(harkio_Log_Info, "calc_impulse done.");

free(tsp_wave_buf);

for (int c = 0; c < input_ch; c++)
    free(iwave[c]);
free(iwave);

IFDEBUG(printf("free2 ..\n"));

IFDEBUG(fclose(debug_fd));

return 0;
}

/*--------------------------------------------------------------------------*/
/* 高速フーリエ変換                                                       */
/*                                                                          */
/*--------------------------------------------------------------------------*/
void fft(double *in_buf, double *out_buf_re, double *out_buf_im, int data_size) {
#ifdef GPL
fftw_complex *in, *out;
fftw_plan p;
int idx;

IFDEBUG(printf("data_size %ld, data_count %ld\n", (long)(sizeof(fftw_complex) * data_size), (long)(data_size)));
in = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
out = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
p = fftw_create_plan(data_size, FFTW_FORWARD, FFTW_ESTIMATE); // plan の作成

// 入力データの設定
for (idx = 0; idx < data_size; idx++) {
    in[idx].re = in_buf[idx];
    in[idx].im = 0;
}

IFDEBUG(printf("calc_fft...\n"));
fftw_one(p, in, out); // fft の実行
IFDEBUG(printf("calced_fft...\n"));

// 出力データの設定
for (idx = 0; idx < data_size; idx++) {
    out_buf_re[idx] = out[idx].re;
    out_buf_im[idx] = out[idx].im;
}

fftw_destroy_plan(p); // plan の破棄
fftw_free(in);
fftw_free(out);

#else
//#endif
//#ifdef HARKLICENSE
/*
 HRI::fft<double> fft_obj(data_size);p

 for (int fft = 0; fft < data_size; fft++)
 {
 fft_obj.a[fft] = std::complex<double>(*(in_buf + fft), 0);
 }

 fft_obj.analyze();pp

 for (int fft = 0; fft < data_size; fft++)
 {
 out_buf_re[fft] = fft_obj.A[fft].real();
 out_buf_im[fft] = fft_obj.A[fft].imag();
 }
 */

int fft;
double _Complex *a;
double _Complex *A;

a = (double _Complex *) malloc(sizeof(double _Complex) * data_size);
A = (double _Complex *) malloc(sizeof(double _Complex) * data_size);

for (fft = 0; fft < data_size; fft++) {
    a[fft] = in_buf[fft] + 0.0f * _Complex_I;
}

//a: srce A:dest
harktool_FFTImpl_dFFT(a, A, data_size);

for (fft = 0; fft < data_size; fft++) {
    out_buf_re[fft] = creal(A[fft]);
    out_buf_im[fft] = cimag(A[fft]);
}

free(a);
free(A);
#endif
}

/*--------------------------------------------------------------------------*/
/*  逆フーリエ高速変換                                                      */
/*                                                                          */
/*--------------------------------------------------------------------------*/
void ifft(double *in_buf_re, double *in_buf_im, double *out_buf_re, double *out_buf_im, int data_size) {
#ifdef GPL
fftw_complex *in, *out;
fftw_plan p;
int idx;

IFDEBUG(printf("data_size %ld, data_count %ld\n", (long)(sizeof(fftw_complex) * data_size), (long) data_size));
in = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
out = (fftw_complex *) fftw_malloc(sizeof(fftw_complex) * data_size);
p = fftw_create_plan(data_size, FFTW_BACKWARD, FFTW_ESTIMATE); // plan の作成

// 入力データの設定
for (idx = 0; idx < data_size; idx++) {
    in[idx].re = in_buf_re[idx];
    in[idx].im = in_buf_im[idx];
}

IFDEBUG(printf("calc_ifft...\n"));
fftw_one(p, in, out); // fft の実行
IFDEBUG(printf("calced_ifft...\n"));

// 出力データの設定
for (idx = 0; idx < data_size; idx++) {
    out_buf_re[idx] = out[idx].re;
    out_buf_im[idx] = out[idx].im;
}

fftw_destroy_plan(p); // plan の破棄
fftw_free(in);
fftw_free(out);
//#endif
//#ifdef HARKLICENSE
#else
/*
 HRI::fft<double> fft_obj(data_size);
 for (int fft = 0; fft < data_size; fft++)
 {
 fft_obj.A[fft] = std::complex<double>(*(in_buf_re+fft), *(in_buf_im+fft));
 }
 fft_obj.synthesize();

 for (int fft = 0; fft < data_size; fft++) {
 out_buf_re[fft] = fft_obj.a[fft].real();
 out_buf_im[fft] = fft_obj.a[fft].imag();
 }
 */

int fft;
double _Complex *A;
double _Complex *a;

A = (double _Complex *) malloc(sizeof(double _Complex) * data_size);
a = (double _Complex *) malloc(sizeof(double _Complex) * data_size);

for (fft = 0; fft < data_size; fft++) {
    A[fft] = in_buf_re[fft] + in_buf_im[fft] * _Complex_I;
}

harktool_FFTImpl_dFFTInv(A, a, data_size);

for (fft = 0; fft < data_size; fft++) {
    out_buf_re[fft] = creal(a[fft]);
    out_buf_im[fft] = cimag(a[fft]);
}

free(A);
free(a);
#endif
}

/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
//void multiply(int n, double *in1_re, double *in1_im, double *in2_re, double *in2_im, double *out_re, double *out_im)
//{
//    for (int i = 0; i < n; i++) {
//        double tmp_re =  in1_re[i] * in2_re[i] - in1_im[i] * in2_im[i];
//        double tmp_im =  in1_re[i] * in2_im[i] + in1_im[i] * in2_re[i];
//        out_re[i] = tmp_re;
//        out_im[i] = tmp_im;
//    }
//}
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
//int corr(double *x, int x_size, double *y, int y_size, int start_pos, int end_pos)
//{
//
//    double  *x_buf;
//    double  *y_buf;
//    double  *cxy;
//    int     idx,idx2,search_len;
//
//    double max_cxy = 0;
//    int    max_idx = 0;
//
//    search_len = end_pos - start_pos;
//
//    x_buf = (double*)malloc(search_len * 2 * sizeof(double));
//    y_buf = (double*)malloc(search_len * 2 * sizeof(double));
//    cxy   = (double*)malloc(search_len * sizeof(double));
//
//    //if (cxy==NULL) {
//    //    fprintf(stderr,"Can't_allocate_memory.\n");
//    //    return 0;
//    //}
//
//    //memset(x_buf, 0, search_len * sizeof(double));
//    memcpy(x_buf, &(x[start_pos]), search_len * sizeof(double));
//    //memset(y_buf, 0, search_len * sizeof(double));
//    memcpy(y_buf, y, search_len * sizeof(double));
//
//    for (idx = 0; idx < search_len-1; idx++) {
//        double s = 0;
//        for (idx2 = 0; idx2 < search_len-1; idx2++) {
//            s += x_buf[idx+idx2] * y_buf[idx2];
//        }
//        cxy[idx] = s;
//
//        if (cxy[idx] > max_cxy) {
//            max_cxy = cxy[idx];
//            max_idx = idx;
//        }
//    }
//
//    printf("searched offset = %d\n", start_pos + max_idx);
//    printf("suggest start offset = %d\n", start_pos + max_idx + DISTANCE_OFFSET);
//
//    free(x_buf);
//    free(y_buf);
//    free(cxy);
//    return start_pos + max_idx + DISTANCE_OFFSET;
//}
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
int corr_on_fft(double *x, int x_size, double *y, int y_size, int start_pos, int end_pos) {
double *x_buf;
//    double  *x_fft_re, *x_fft_im;
COMPLEX_DOUBLE* x_fft;
double *y_buf;
//    double  *y_fft_re, *y_fft_im;
COMPLEX_DOUBLE* y_fft;
//    double  *fft_res_re, *fft_res_im;
COMPLEX_DOUBLE* fft_res;
//    double  *res_buf_re, *res_buf_im;
COMPLEX_DOUBLE* res_buf;
double *cxy;
int peak_corr_idx[10];
int search_len, len_temp;

double max_cxy = 0;
int max_idx = 0;

int fft_len_log, fft_len;

search_len = end_pos - start_pos;

if (search_len < y_size) {
    len_temp = y_size;
} else {
    len_temp = search_len;
}

fft_len_log = (int) floor(log(len_temp) / log(2));

fft_len = (int) pow(2, fft_len_log); /* for fft */

x_buf = (double*) malloc(search_len * sizeof(double));
//    x_fft_re   = (double*)malloc(fft_len * sizeof(double));
//    x_fft_im   = (double*)malloc(fft_len * sizeof(double));
x_fft = (COMPLEX_DOUBLE*) malloc(fft_len * sizeof(COMPLEX_DOUBLE));
y_buf = (double*) malloc(y_size * sizeof(double));
//    y_fft_re   = (double*)malloc(fft_len * sizeof(double));
//    y_fft_im   = (double*)malloc(fft_len * sizeof(double));
y_fft = (COMPLEX_DOUBLE*) malloc(fft_len * sizeof(COMPLEX_DOUBLE));
//    fft_res_re = (double*)malloc(fft_len * sizeof(double));
//    fft_res_im = (double*)malloc(fft_len * sizeof(double));
fft_res = (COMPLEX_DOUBLE*) malloc(fft_len * sizeof(COMPLEX_DOUBLE));
//    res_buf_re = (double*)malloc(fft_len * sizeof(double));
//    res_buf_im = (double*)malloc(fft_len * sizeof(double));
res_buf = (COMPLEX_DOUBLE*) malloc(fft_len * sizeof(COMPLEX_DOUBLE));
cxy = (double*) malloc(fft_len * sizeof(double));

if ((x_buf == NULL) || (x_fft == NULL) || (y_buf == NULL) || (y_fft == NULL) || (fft_res == NULL) || (res_buf == NULL)
        || (cxy == NULL)) {
    fprintf(stderr, "Can't_allocate_memory.\n");
    return 0;
}

//memset(x_buf, 0, search_len * sizeof(double));
//    memset(x_fft_re, 0, fft_len * sizeof(double));
//    memset(x_fft_im, 0, fft_len * sizeof(double));
memset(x_fft, 0, fft_len * sizeof(COMPLEX_DOUBLE));
memcpy(x_buf, &(x[start_pos]), search_len * sizeof(double));
//memset(y_buf, 0, y_size * sizeof(double));
//    memset(y_fft_re, 0, fft_len * sizeof(double));
//    memset(y_fft_im, 0, fft_len * sizeof(double));
memset(y_fft, 0, fft_len * sizeof(COMPLEX_DOUBLE));
memcpy(y_buf, y, y_size * sizeof(double));

//    memset(fft_res_re, 0, fft_len * sizeof(double));
//    memset(fft_res_im, 0, fft_len * sizeof(double));
memset(fft_res, 0, fft_len * sizeof(COMPLEX_DOUBLE));
//    memset(res_buf_re, 0, fft_len * sizeof(double));
//    memset(res_buf_im, 0, fft_len * sizeof(double));
memset(res_buf, 0, fft_len * sizeof(COMPLEX_DOUBLE));

memset(peak_corr_idx, 0, 10);

//    fft(x_buf, x_fft_re, x_fft_im, fft_len);
harktool_FFTImpl_dRealFFT(x_buf, x_fft, fft_len);
harktool_FFTImpl_dRealFFT(y_buf, y_fft, fft_len);

free(x_buf);
free(y_buf);

int idx;
int corr_idx;

for (idx = 0; idx < fft_len; idx++) {
    y_fft[idx] = conj(y_fft[idx]);
    fft_res[idx] = y_fft[idx] * x_fft[idx];
}

//    multiply(fft_len, y_fft_re, y_fft_im, x_fft_re, x_fft_im, fft_res_re, fft_res_im);

//    ifft(fft_res_re, fft_res_im, res_buf_re, res_buf_im, fft_len);
harktool_FFTImpl_dFFTInv(fft_res, res_buf, fft_len);

for (idx = 0; idx < fft_len; idx++) {

//        cxy[idx]= sqrt((res_buf_re[idx] * res_buf_re[idx]) + (res_buf_im[idx] * res_buf_im[idx]));
    cxy[idx] = cabs(res_buf[idx]);

    if (cxy[idx] > max_cxy) {
        max_cxy = cxy[idx];
        max_idx = idx;

        for (corr_idx = 9; corr_idx > 0; corr_idx--) {
            peak_corr_idx[corr_idx] = peak_corr_idx[corr_idx - 1];
        }
        peak_corr_idx[0] = max_idx;
    }
}

for (corr_idx = 0; corr_idx < 10; corr_idx++) {
    IFDEBUG(printf("10 peak %d:%d:%f\n", corr_idx, peak_corr_idx[corr_idx], cxy[peak_corr_idx[corr_idx]]));
    if ((peak_corr_idx[corr_idx] < (max_idx - (16384 / 4) * 3))
            && (cxy[peak_corr_idx[corr_idx]] > (cxy[max_idx] / 2))) {
        IFDEBUG(printf("max_idx,peak_idx %d:%d\n", max_idx, peak_corr_idx[corr_idx]));
        max_idx = peak_corr_idx[corr_idx];
    }
    if (peak_corr_idx[corr_idx] == 0) {
//            free(x_fft_re);
//            free(x_fft_im);
//            free(y_fft_re);
//            free(y_fft_im);
//            free(fft_res_re);
//            free(fft_res_im);
//            free(res_buf_re);
//            free(res_buf_im);
        free(x_fft);
        free(y_fft);
        free(fft_res);
        free(res_buf);
        free(cxy);
        return 0;
    }
}

printf("searched offset = %d\n", start_pos + max_idx);
printf("suggest start offset = %d\n", start_pos + max_idx + DISTANCE_OFFSET);

//    free(x_fft_re);
//    free(x_fft_im);
//    free(y_fft_re);
//    free(y_fft_im);
//    free(fft_res_re);
//    free(fft_res_im);
//    free(res_buf_re);
//    free(res_buf_im);
free(x_fft);
free(y_fft);
free(fft_res);
free(res_buf);
free(cxy);

return start_pos + max_idx + DISTANCE_OFFSET;
}
